# 引入哈希库
import hashlib
import random
import datetime
from .sqlalchemy import BaseModel, Base
from sqlalchemy import Column, Integer, String, DATETIME
from .config import config


class UrlUuid(Base, BaseModel):
    # 表的名字:
    __tablename__ = 'short_urls'
    __label__ = 'mysql_short_url'
    __database__ = config(__label__, 'database')
    id = Column(String(16), primary_key=True)
    url = Column(String(128))
    create_time = Column(DATETIME())
    modify_time = Column(DATETIME())


def get_md5(s):
    s = s.encode('utf8')
    m = hashlib.md5()
    print(m)
    m.update(s)
    return m.hexdigest()


def get_sha224(s):
    code = hashlib.sha3_256(s.encode('utf8'))
    return code.hexdigest()


code_map = (
    'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h',
    'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p',
    'q', 'r', 's', 't', 'u', 'v', 'w', 'x',
    'y', 'z', '0', '1', '2', '3', '4', '5',
    '6', '7', '8', '9', 'A', 'B', 'C', 'D',
    'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L',
    'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T',
    'U', 'V', 'W', 'X', 'Y', 'Z'
)


def get_hash_key(long_url):
    hkeys = []
    hex = get_sha224(long_url)
    for i in range(0, 8):
        n = int(hex[i * 8:(i + 1) * 8], 16)
        v = []
        e = 0
        for j in range(0, 7):
            x = 0x0000003D & n
            e |= ((0x00000002 & n) >> 1) << j
            v.insert(0, code_map[x])
            n = n >> 6
        e |= n << 5
        v.insert(0, code_map[e & 0x0000003D])
        hkeys.append(''.join(v))
    return hkeys


def get_url_uuid_list(long_url):
    uuid_list = get_hash_key(long_url)
    return uuid_list


def insert_url_hash_key(long_url):
    uuid_list = get_url_uuid_list(long_url)
    _res = {
        'code': -1,  # 0插入成功，1已插入，-1失败
        'msg': 'error',
        'uuid': ''
    }
    res = UrlUuid.get_by_ids(uuid_list)
    _now = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')
    for url_entity in res:
        if url_entity.url == long_url:
            _res['uuid'], _res['code'], _res['msg'] = url_entity.id, 1, 'success'
            return _res
        else:
            uuid_list.remove(url_entity.id)
    if len(uuid_list) == 0:
        _res['msg'] = 'not enough uuid'
    else:
        uuid = uuid_list[random.randint(0, len(uuid_list))]
        url_entity = UrlUuid(id=uuid, url=long_url, create_time=_now, modify_time=_now)
        UrlUuid.save(url_entity)
        _res['uuid'], _res['code'], _res['msg'] = uuid, 0, 'success'
    return _res



# coding: utf-8
import os
import sys
import string
from shlex import shlex

# Useful for very coarse version differentiation.
PY3 = sys.version_info[0] == 3

if PY3:
    from configparser import ConfigParser

    text_type = str
else:
    from ConfigParser import SafeConfigParser as ConfigParser

    text_type = unicode


class UndefinedValueError(Exception):
    pass


class Undefined(object):
    """
    Class to represent undefined type.
    """
    pass


# Reference instance to represent undefined values
undefined = Undefined()


class Config(object):
    """
    Handle .env file format used by Foreman.
    """
    _BOOLEANS = {'1': True, 'yes': True, 'true': True, 'on': True,
                 '0': False, 'no': False, 'false': False, 'off': False, '': False}

    def __init__(self, repository):
        self.repository = repository

    def _cast_boolean(self, value):
        """
        Helper to convert config values to boolean as ConfigParser do.
        """
        value = str(value)
        if value.lower() not in self._BOOLEANS:
            raise ValueError('Not a boolean: %s' % value)

        return self._BOOLEANS[value.lower()]

    @staticmethod
    def _cast_do_nothing(value):
        return value

    def get(self, section, option=None, default=None, cast=undefined):
        """
        Return the value for option or default if defined.
        """
        if option is None:
            option = section
            section = 'settings'

        value = self.repository.get(section, option,default)

        if isinstance(cast, Undefined):
            cast = self._cast_do_nothing
        elif cast is bool:
            cast = self._cast_boolean

        return cast(value)

    def __call__(self, *args, **kwargs):
        """
        Convenient shortcut to get.
        """
        return self.get(*args, **kwargs)


class RepositoryEmpty(object):
    def __init__(self, source=''):
        pass

    def __contains__(self, key):
        return False

    def __getitem__(self, key):
        return None


class RepositoryIni(RepositoryEmpty):
    """
    Retrieves option keys from .ini files.
    """
    SECTION = 'settings'

    def __init__(self, source):
        self.parser = ConfigParser()
        with open(source) as file_:
            self.parser.read_file(file_)

    def get(self, section, key, default_value=None):
        if default_value is not None:
            try:
                return self.parser.get(section, key)
            except:
                return default_value
        return self.parser.get(section, key)


class RepositoryEnv(RepositoryEmpty):
    """
    Retrieves option keys from .env files with fall back to os.environ.
    """

    def __init__(self, source):
        self.data = {}

        with open(source) as file_:
            for line in file_:
                line = line.strip()
                if not line or line.startswith('#') or '=' not in line:
                    continue
                k, v = line.split('=', 1)
                k = k.strip()
                v = v.strip().strip('\'"')
                self.data[k] = v

    def __contains__(self, key):
        return key in os.environ or key in self.data

    def get(self, section='', key=''):
        return self.data[key]


class AutoConfig(object):
    """
    Autodetects the config file and type.

    Parameters
    ----------
    search_path : str, optional
        Initial search path. If empty, the default search path is the
        caller's path.

    """
    SUPPORTED = {
        'settings.ini': RepositoryIni,
        '.env': RepositoryEnv,
    }

    def __init__(self, search_path=None):
        self.search_path = search_path
        self.config = None

    def _find_file(self, path):
        # look for all files in the current path
        for configfile in self.SUPPORTED:
            filename = os.path.join(path, configfile)
            if os.path.isfile(filename):
                return filename

        # search the parent
        parent = os.path.dirname(path)
        if parent and parent != os.path.sep:
            return self._find_file(parent)

        # reached root without finding any files.
        return ''

    def _load(self, path):
        # Avoid unintended permission errors
        try:
            filename = self._find_file(os.path.abspath(path))
        except Exception:
            filename = ''
        Repository = self.SUPPORTED.get(os.path.basename(filename), RepositoryEmpty)

        self.config = Config(Repository(filename))

    def _caller_path(self):
        # MAGIC! Get the caller's module path.
        frame = sys._getframe()
        path = os.path.dirname(frame.f_back.f_back.f_code.co_filename)
        return path

    def __call__(self, *args, **kwargs):
        if not self.config:
            self._load(self.search_path or self._caller_path())

        return self.config(*args, **kwargs)


# A pré-instantiated AutoConfig to improve decouple's usability
# now just import config and start using with no configuration.
config = AutoConfig()


# Helpers

class Csv(object):
    """
    Produces a csv parser that return a list of transformed elements.
    """

    def __init__(self, cast=text_type, delimiter=',', strip=string.whitespace, post_process=list):
        """
        Parameters:
        cast -- callable that transforms the item just before it's added to the list.
        delimiter -- string of delimiters chars passed to shlex.
        strip -- string of non-relevant characters to be passed to str.strip after the split.
        tuple_ -- boolean to check if it is to return in tuple format.
        """
        self.cast = cast
        self.delimiter = delimiter
        self.strip = strip
        self.post_process = post_process

    def __call__(self, value):
        """The actual transformation"""
        transform = lambda s: self.cast(s.strip(self.strip))

        splitter = shlex(value, posix=True)
        splitter.whitespace = self.delimiter
        splitter.whitespace_split = True

        return self.post_process(transform(s) for s in splitter)



import multiprocessing
from .log import *
from .utils import *


class MultiWorker(multiprocessing.Process):

    def __init__(self, jobs, index=0, obj=None):
        multiprocessing.Process.__init__(self)
        self.jobs = jobs
        self.index = index
        self.obj = obj

    def run(self):
        log_info("start run job work size=%d index=%d" % (len(self.jobs), self.index))
        for job in self.jobs:
            self.handle(job)
        log_info("done run job work size=%d index=%d" % (len(self.jobs), self.index))

    def handle(self, job):
        raise Exception('need impl')


class MultiWorkExecutor(object):

    def __init__(self, worker_class, jobs, work_num, name, obj=None):
        self.jobs = jobs
        self.work_num = work_num
        self.worker_class = worker_class
        self.name = name
        self.obj = obj

    def execute(self):
        log_info(
            "start multiprocessing handle job size=%d,work_num=%d for %s" % (len(self.jobs), self.work_num, self.name))
        start = get_current_second()
        if self.work_num <= 1 or len(self.jobs) <= self.work_num:
            worker = self.worker_class(self.jobs)
            worker.start()
            return
        items = list()
        for i in range(self.work_num):
            items.append(list())
        for index, job in enumerate(self.jobs):
            items[index % self.work_num].append(job)
        workers = list()
        for index, jobs in enumerate(items):
            worker = self.worker_class(jobs, index, self.obj)
            workers.append(worker)
        for worker in workers:
            worker.start()
        for worker in workers:
            worker.join()
        log_info(
            "end multiprocessing handle job size=%d,work_num=%d for %s cost=%d" % (
                len(self.jobs), self.work_num, self.name, (get_current_second() - start)))



from .queue0 import *
from .log import log_exception


class RabbitProducer(RabbitProducer0):

    def send_safely(self, topic, message):
        try:
            self.send(topic, message)
            return True
        except Exception as ex:
            log_exception(str(ex))
            self.close_safely()
            return False

    def close_safely(self):
        try:
            self.close()
        except Exception as ex:
            log_exception(str(ex))
